package com.nimbusds.openid.connect.provider.userinfo.ldap;


import java.util.Properties;

import org.apache.log4j.Logger;

import com.unboundid.ldap.sdk.DN;
import com.unboundid.ldap.sdk.LDAPException;
import com.unboundid.ldap.sdk.SearchScope;

import com.thetransactioncompany.util.PropertyParseException;
import com.thetransactioncompany.util.PropertyRetriever;

import com.nimbusds.common.config.ConfigurationException;
import com.nimbusds.common.config.CustomKeyStoreConfiguration;
import com.nimbusds.common.config.CustomTrustStoreConfiguration;
import com.nimbusds.common.config.DirectoryUser;
import com.nimbusds.common.config.LDAPServerConnectionPoolDetails;
import com.nimbusds.common.config.LoggableConfiguration;

import com.nimbusds.common.ldap.FilterTemplate;


/**
 * UserInfo LDAP connector configuration. It is typically derived from a Java
 * key / value properties file. The configuration is stored as public fields
 * which become immutable (final) after their initialisation.
 *
 * <p>Example configuration properties:
 *
 * <pre>
 * op.userInfo.ldap.server.url = ldap://localhost:1389 ldap://remotehost:1389
 * op.userInfo.ldap.server.selectionAlgorithm = FAILOVER
 * op.userInfo.ldap.server.connectTimeout = 1000
 * op.userInfo.ldap.server.security = STARTTLS
 * op.userInfo.ldap.server.trustSelfSignedCerts = true
 * op.userInfo.ldap.server.connectionPoolSize = 10
 * op.userInfo.ldap.server.connectionPoolMaxWaitTime = 100
 *
 * op.userInfo.ldap.directory.user.dn = cn=Directory Manager
 * op.userInfo.ldap.directory.user.password = secret
 *
 * op.userInfo.ldap.directory.baseDN = ou=people,dc=wonderland,dc=net
 * op.userInfo.ldap.directory.scope = ONE
 * op.userInfo.ldap.directory.filter = (uid=%u)
 *
 * op.userInfo.ldap.directory.attributeMap = oidcUserInfoMap.json
 *
 * op.userInfo.ldap.customTrustStore.enable = false
 * op.userInfo.ldap.customTrustStore.file =
 * op.userInfo.ldap.customTrustStore.password =
 * op.userInfo.ldap.customTrustStore.type =
 *
 * op.userInfo.ldap.customKeyStore.enable = false
 * op.userInfo.ldap.customKeyStore.file =
 * op.userInfo.ldap.customKeyStore.password =
 * op.userInfo.ldap.customKeyStore.type =
 * </pre>
 *
 * @author Vladimir Dzhuvinov
 */
public final class Configuration implements LoggableConfiguration {


	/**
	 * The default properties prefix.
	 */
	public static final String DEFAULT_PREFIX = "op.userInfo.ldap.";


	/**
	 * Directory store details.
	 */
	public static class Directory implements LoggableConfiguration {


		/**
		 * The directory user credentials. Should have read permission
		 * to the directory tree containing the user entries.
		 *
		 * <p>Property keys: [prefix]user.*
		 */
		public final DirectoryUser user;


		/**
		 * The base distinguished name (DN) of the directory branch
		 * where the user entries are stored.
		 *
		 * <p>Property key: [prefix]baseDN
		 */
		public final DN baseDN;


		/**
		 * The search scope for the user entries with respect to the
		 * specified base DN.
		 *
		 * <p>Property key: [prefix]scope
		 */
		public final SearchScope scope;


		/**
		 * The search filter to retrieve the user entries. The "%u"
		 * placeholder will be substituted with the subject identifier.
		 *
		 * <p>Property key: [prefix]filter
		 */
		public final FilterTemplate filter;


		/**
		 * The file name containing the LDAP attribute / JSON object
		 * member map.
		 *
		 * <p>Property key: [prefix]attributeMap
		 */
		public final String attributeMap;


		/**
		 * Creates a new directory store details instance from the
		 * specified properties.
		 *
		 * @param prefix The properties prefix. Must not be
		 *               {@code null}.
		 * @param props  The properties. Must not be {@code null}.
		 *
		 * @throws PropertyParseException On a missing or invalid
		 *                                property.
		 */
		public Directory(final String prefix, final Properties props)
			throws PropertyParseException {

			user = new DirectoryUser(prefix + "user.", props);

			PropertyRetriever pr = new PropertyRetriever(props);

			String dnString = pr.getString(prefix + "baseDN");

			try {
				baseDN = new DN(dnString);

			} catch (LDAPException e) {

				throw new PropertyParseException("Invalid base DN: " + e.getMessage(),
					prefix + "baseDN",
					dnString);
			}

			String scopeString = pr.getString(prefix + "scope");

			if (scopeString.equalsIgnoreCase("BASE"))
				scope = SearchScope.BASE;
			else if (scopeString.equalsIgnoreCase("ONE"))
				scope = SearchScope.ONE;
			else if (scopeString.equalsIgnoreCase("SUB"))
				scope = SearchScope.SUB;
			else if (scopeString.equalsIgnoreCase("SUBORDINATES"))
				scope = SearchScope.SUBORDINATE_SUBTREE;
			else
				throw new PropertyParseException("Invalid search scope",
					prefix + "scope",
					scopeString);

			String filterString = pr.getString(prefix + "filter");

			try {
				filter = new FilterTemplate(filterString);

			} catch (IllegalArgumentException e) {

				throw new PropertyParseException("Invalid filter template: " + e.getMessage(),
					prefix + "filter",
					filterString);
			}

			attributeMap = pr.getString(prefix + "attributeMap");
		}


		@Override
		public void log() {

			Logger log = Logger.getLogger(Directory.class);

			user.log();

			log.info("UserInfo LDAP connector: Base DN: " + baseDN.toString());
			log.info("UserInfo LDAP connector: Search scope: " + scope);
			log.info("UserInfo LDAP connector: Filter: " + filter);
			log.info("UserInfo LDAP connector: Attribute map file: " + attributeMap);
		}
	}


	/**
	 * The LDAP server connect details.
	 */
	public final LDAPServerConnectionPoolDetails server;


	/**
	 * The directory store details.
	 */
	public final Directory directory;


	/**
	 * The custom trust store details.
	 */
	public final CustomTrustStoreConfiguration customTrustStore;


	/**
	 * The custom key store details.
	 */
	public final CustomKeyStoreConfiguration customKeyStore;


	/**
	 * Creates a new UserInfo configuration from the specified properties.
	 *
	 * @param props The properties. Must not be {@code null}.
	 *
	 * @throws Configuration exception On a missing or invalid property.
	 */
	public Configuration(final Properties props)
		throws ConfigurationException {

		PropertyRetriever pr = new PropertyRetriever(props);

		try {
			server = new LDAPServerConnectionPoolDetails(DEFAULT_PREFIX + "server.", props);

			directory = new Directory(DEFAULT_PREFIX + "directory.", props);

			customTrustStore = new CustomTrustStoreConfiguration(DEFAULT_PREFIX + "customTrustStore.", props);

			customKeyStore = new CustomKeyStoreConfiguration(DEFAULT_PREFIX + "customKeyStore.", props);

		} catch (PropertyParseException e) {

			throw new ConfigurationException(e.getMessage() +
				": Property: " + e.getPropertyKey());
		}
	}


	/**
	 * Logs the configuration details at INFO level. Properties that may
	 * adversely affect security are logged at WARN level.
	 */
	@Override
	public void log() {

		Logger log = Logger.getLogger(Configuration.class);

		log.info("UserInfo LDAP connector configuration:");

		server.log();
		directory.log();
		customTrustStore.log();
		customKeyStore.log();
	}
}

